# Copyright 2023 Eurobios
# Licensed under the CeCILL License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     https://cecill.info/
import numpy as np
import pandas as pd


def build_admissible_data(df: pd.DataFrame, id_discrete: iter, id_continuous: iter) -> pd.DataFrame:
    df.index = range(len(df.index))
    df_non_admissible = sample_non_admissible_data(df, id_discrete, id_continuous).__deepcopy__()
    df_non_admissible = create_zones(df_non_admissible, id_discrete, id_continuous)

    df_non_admissible["__id__"] = range(len(df_non_admissible))
    df_add_non_admissible = df_non_admissible[[*id_discrete, *id_continuous, "__id__", "__zone__"]]
    all_id_continuous = df_non_admissible[id_continuous[0]].to_list()
    all_id_continuous += df_non_admissible[id_continuous[1]].to_list()

    # create good segmentation
    df_ret = pd.concat([df_non_admissible[[*id_discrete, "__zone__"]]] * 2)
    df_ret[id_continuous[0]] = all_id_continuous
    df_ret["__disc__"] = compute_discontinuity(df_ret, id_discrete, id_continuous)
    df_ret = df_ret.sort_values(by=[*id_discrete, id_continuous[0]])
    df_ret[id_continuous[1]] = - df_ret[id_continuous[0]].diff(periods=-1) + df_ret[id_continuous[0]]
    df_ret = df_ret.dropna().drop(columns="__disc__")
    df_ret = df_ret.drop_duplicates().dropna()
    df_ret = df_ret[df_ret[id_continuous[1]] != df_ret[id_continuous[0]]]
    df_ret = df_ret.sort_values(by=[*id_discrete, id_continuous[0]])
    df_ret[id_continuous[1]] = df_ret[id_continuous[1]].astype(df[id_continuous[1]].dtype)

    df_ret = pd.merge(df_ret, df_add_non_admissible,
                      on=id_discrete + ["__zone__"], suffixes=("", "_tmp"))
    id_continuous_tmp = [str(i) + "_tmp" for i in id_continuous]
    c = df_ret[id_continuous[0]] < df_ret[id_continuous_tmp[1]]
    c &= df_ret[id_continuous[1]] > df_ret[id_continuous_tmp[0]]
    df_ret = df_ret.loc[c].drop(columns=id_continuous_tmp)

    df_ret = pd.merge(df_ret, df_non_admissible.drop(columns=[*id_discrete, *id_continuous, "__zone__"]), on="__id__"
                      ).drop(columns=["__id__", "__zone__"])
    df_ret = df_ret.astype(df_non_admissible.dtypes.drop(["__id__", "__zone__"]))

    df_ret_all = df[~get_overlapping(df, id_discrete, id_continuous)]
    df_ret_all = pd.concat((df_ret_all, df_ret))
    df_ret_all = df_ret_all.sort_values(by=[*id_discrete, id_continuous[0]])
    df_ret_all.index = range(len(df_ret_all.index))
    return df_ret_all


def create_zones(df: pd.DataFrame, id_discrete: iter, id_continuous: iter):
    """
    Create overlapping zone identifiers in the DataFrame based on discrete and continuous ID columns.

    Parameters
    ----------
    df : pd.DataFrame
        The input DataFrame containing the df.
    id_discrete : iter
        An iterable of column names that are considered discrete identifiers.
    id_continuous : iter
        An iterable of column names that are considered continuous identifiers.

    Returns
    -------
    pd.DataFrame
        The DataFrame with an additional '__zone__' column indicating the zone for each row.

    Notes
    -----
    The function works by sorting the DataFrame based on the given discrete and continuous identifiers,
    and then creating a zone identifier (`__zone__`) that groups rows based on specific conditions.

    Steps:
    1. Sort the DataFrame based on discrete identifiers and the second continuous identifier.
    2. Assign a forward index (`__zf__`) based on the sorted order.
    3. Sort the DataFrame based on discrete identifiers and the first continuous identifier.
    4. Assign a backward index (`__zi__`) based on the sorted order.
    5. Determine zones where the forward and backward indices are equal (`c_zone`).
    6. Check if the start of a zone is greater than or equal to the end of the previous zone (`c_inner`).
    7. Identify changes in discrete identifiers (`c_disc`).
    8. Combine the conditions to create the final zone identifier (`__zone__`).
    9. Merge the zone information back into the original DataFrame.

    Examples
    --------
    >>> df = {
    ...     'id': [1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2],
    ...     't1': [932, 996, 2395, 2395, 3033, 3628, 4126, 4140, 4154, 316263, 316263, 316471, 316471],
    ...     't2': [2395, 2324, 3033, 3628, 3035, 4140, 4140, 5508, 5354, 316399, 316471, 317406, 317557],
    ...     'LONGUEUR': [1463, 1328, 638, 1233, 2, 512, 14, 1368, 1200, 136, 208, 935, 1086],
    ...     '__zone__': [0, 0, 1, 1, 1, 2, 2, 2, 2, 3, 3, 4, 4]
    ... }
    >>> df = pd.DataFrame(df)
    >>> create_zones(df, ['id'], ['t1', 't2'])
    """
    df_out = df.__deepcopy__()
    if "__zone__" in df.columns:
        df_out = df_out.drop(columns='__zone__')

    df_out = df_out.sort_values([*id_discrete, id_continuous[1]])
    df_out["__zf__"] = range(len(df_out))
    df_out = df_out.sort_values([*id_discrete, id_continuous[0]])
    df_out["__zi__"] = range(len(df_out))
    c_zone = (df_out["__zf__"] - df_out["__zi__"]) == 0

    df_out["__id2_prev__"] = df[id_continuous[1]]
    df_out.loc[df_out.index[1:], "__id2_prev__"] = df_out.loc[df_out.index[:-1], id_continuous[1]].values
    c_inner = df_out[id_continuous[0]] >= df_out["__id2_prev__"]

    c_disc = df_out[id_discrete].iloc[1:].values == df_out[id_discrete].iloc[:-1].values
    c_disc = c_disc.mean(axis=1) == 1
    c_disc = ~ np.concatenate(([True], c_disc))

    df_out["__zone__"] = (c_zone & c_inner) | c_disc
    df_out["__zone__"] = df_out["__zone__"].cumsum()

    return df_out.loc[df.index, [*df.columns.to_list(), "__zone__"]]


def get_overlapping(df: pd.DataFrame,
                    id_discrete: iter,
                    id_continuous: iter) -> pd.Series:
    df = create_zones(df, id_discrete, id_continuous)
    overlap = df["__zone__"].duplicated(keep=False)
    return overlap


def admissible_dataframe(data: pd.DataFrame,
                         id_discrete: iter,
                         id_continuous: iter):
    return sum(get_overlapping(data, id_discrete,
                               id_continuous)) == 0


def sample_non_admissible_data(data: pd.DataFrame,
                               id_discrete: iter,
                               id_continuous: iter) -> pd.DataFrame:
    return data[get_overlapping(data, id_discrete,
                                id_continuous)]


def compute_discontinuity(df, id_discrete, id_continuous):
    """
    Compute discontinuity in rail segment. The i-th element in return
    will be True if i-1 and i are discontinuous

    """
    discontinuity = np.zeros(len(df)).astype(bool)
    for col in id_discrete:
        if col in df.columns:
            discontinuity_temp = np.concatenate(
                ([False], df[col].values[1:] != df[col].values[:-1]))
            discontinuity |= discontinuity_temp

    if id_continuous[0] in df.columns and id_continuous[1] in df.columns:
        discontinuity_temp = np.concatenate(
            ([False], df[id_continuous[0]].values[1:] != df[id_continuous[
                1]].values[:-1]))
        discontinuity |= discontinuity_temp
    return discontinuity



# Copyright 2023 Eurobios
# Licensed under the CeCILL License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     https://cecill.info/
import numpy as np
import pandas as pd
from crep import tools


def merge(
        data_left: pd.DataFrame,
        data_right: pd.DataFrame,
        id_continuous: iter,
        id_discrete: iter, how: str,
        remove_duplicates: bool = False,
        verbose=False) -> pd.DataFrame:
    """
    This function aims at creating merge data frame

    Parameters
    ----------

    data_left
        data frame with continuous representation
    data_right
        data frame with continuous representation
    id_continuous
        iterable of length two that delimits the edges of the segment
    id_discrete
        iterable that lists all the columns on which to perform a classic merge
    how
        how to make the merge, possible options are

        - 'left'
        - 'right'
        - 'inner'
        - 'outer'

    remove_duplicates
        whether to remove duplicates
    verbose
    """
    __check_args_merge(data_left, data_right,
                       id_continuous, id_discrete, how)

    data_left = data_left.__deepcopy__()
    data_right = data_right.__deepcopy__()
    id_continuous = list(id_continuous)
    id_discrete = list(id_discrete)

    id_discrete_left = [col for col in data_left.columns if col in id_discrete]
    id_discrete_right = [col for col in data_right.columns if col in id_discrete]
    data_left, data_right = __fix_discrete_index(
        data_left, data_right,
        id_discrete_left,
        id_discrete_right)
    data_left.index = range(len(data_left))
    data_right.index = range(len(data_right))
    df_merge = __merge_index(data_left, data_right,
                             id_discrete=id_discrete,
                             id_continuous=id_continuous)
    df = pd.merge(
        df_merge,
        data_left[list(set(data_left.columns).difference(df_merge.columns))],
        left_on="left_idx", right_index=True, how="left")
    df = pd.merge(
        df,
        data_right[list(set(data_right.columns).difference(df_merge.columns))],
        left_on="right_idx", right_index=True, how="left")

    if how == "left":
        df = df.loc[df["left_idx"] != -1]
    if how == "right":
        df = df.loc[df["right_idx"] != -1]
    if how == "inner":
        df = df.loc[(df["right_idx"] != -1) & (df["left_idx"] != -1)]

    df = df.drop(["left_idx", "right_idx"], axis=1)
    df.index = range(len(df))
    if remove_duplicates:
        df = suppress_duplicates(df, id_discrete=id_discrete,
                                 continuous_index=id_continuous)
    if verbose:
        print("[merge] nb rows left  table frame ", data_left.shape[0])
        print("[merge] nb rows right table frame ", data_right.shape[0])
        print("[merge] nb rows outer table frame ", df.shape[0])
    return df


def unbalanced_merge(data_admissible: pd.DataFrame,
                     data_not_admissible: pd.DataFrame, id_discrete, id_continuous) -> pd.DataFrame:
    """
    Merge admissible and non-admissible dataframes based on discrete and continuous identifiers.

    Parameters
    ----------
    data_admissible : pd.DataFrame
        DataFrame containing admissible data.
    data_not_admissible : pd.DataFrame
        DataFrame containing non-admissible data.
    id_discrete : list
        List of column names representing discrete identifiers.
    id_continuous : list
        List of column names representing continuous identifiers.

    Returns
    -------
    pd.DataFrame
        A DataFrame resulting from the unbalanced merge of admissible and non-admissible data.

    Notes
    -----
    The function performs the following steps:
    1. Combines and sorts the admissible and non-admissible data based on the identifiers.
    2. Resolves overlaps and conflicts between the admissible and non-admissible data.
    3. Merges and returns the final DataFrame.
    """
    # assert tools.admissible_dataframe(data_admissible, id_discrete, id_continuous)
    df_idx_w = data_admissible[[*id_discrete, *id_continuous]].copy()
    df_idx_s = data_not_admissible[[*id_discrete, *id_continuous]].copy()
    df_idx_w["__t__"] = True
    df_idx_s["__t__"] = False

    df_idx = pd.concat((df_idx_w, df_idx_s))
    df_idx = df_idx.sort_values([*id_discrete, id_continuous[0], "__t__"],
                                ascending=[*[True] * len(id_discrete), True, False])

    df_idx["__id2__"] = np.nan
    df_idx["__id1__"] = np.nan
    df_idx.loc[df_idx["__t__"], "__id2__"] = df_idx.loc[df_idx["__t__"], id_continuous[1]]
    df_idx.loc[df_idx["__t__"], "__id1__"] = df_idx.loc[df_idx["__t__"], id_continuous[0]]
    df_idx["__id2__"] = df_idx["__id2__"].ffill()
    df_idx["__id1__"] = df_idx["__id1__"].ffill()

    c_resolve = df_idx["__id2__"] < df_idx[id_continuous[1]]
    c_out = df_idx["__id2__"] < df_idx[id_continuous[0]]
    created_columns = ["__t__", "__id2__", "__id1__"]

    # =================
    # Encompassed data
    df_admissible = df_idx[~c_resolve].copy()
    df_admissible_ret = pd.merge(df_admissible, data_admissible, how='inner',
                                 left_on=[*id_discrete, "__id1__", "__id2__"], right_on=[*id_discrete, *id_continuous],
                                 suffixes=("", "__init"))
    df_admissible_ret = df_admissible_ret[~df_admissible_ret.__t__]
    df_admissible_ret = df_admissible_ret.drop(
        columns=[*created_columns, id_continuous[0] + "__init", id_continuous[1] + "__init"])
    df_admissible_ret = pd.merge(df_admissible_ret, data_not_admissible, on=[*id_discrete, *id_continuous], how="left")

    # =================
    # To resolve data
    df_to_resolve = df_idx[c_resolve & ~c_out].copy()
    old = [f"{id_continuous[0]}__", f"{id_continuous[1]}__"]
    df_to_resolve[old] = df_to_resolve[id_continuous]
    df_to_resolve_admissible = tools.build_admissible_data(df_to_resolve.drop(columns=created_columns), id_discrete,
                                                     id_continuous)

    df_to_resolve_no_d = df_to_resolve_admissible.drop_duplicates(subset=[*id_discrete, *id_continuous])
    df_to_resolve_no_d = merge(df_to_resolve_no_d, data_admissible, id_discrete=id_discrete, id_continuous=id_continuous,
                               how="inner")
    df_ret = pd.merge(
        df_to_resolve_no_d,
        data_not_admissible,
        left_on=[*id_discrete, *old],
        right_on=[*id_discrete, *id_continuous], how="inner", suffixes=("", "__")
    )
    df_ret = df_ret[df_admissible_ret.columns]

    # =================
    # Out data
    df_to_out = df_idx[c_resolve & c_out].copy().drop(columns=created_columns)
    df_to_out = pd.merge(df_to_out, data_not_admissible, on=[*id_discrete, *id_continuous], how='inner')

    df_ret_all = pd.concat((df_ret, df_admissible_ret, df_to_out), axis=0)
    df_ret_all.index = range(len(df_ret_all))
    df_ret_all = df_ret_all.sort_values(by=[*id_discrete, *id_continuous])
    return df_ret_all


def aggregate_constant(df: pd.DataFrame,
                       id_discrete: iter,
                       id_continuous: iter,
                       ):
    """

    Parameters
    ----------
    df
    id_discrete
    id_continuous

    Returns
    -------

    """
    data_ = df.copy(deep=True)
    dtypes = data_.dtypes
    data_ = data_.sort_values([*id_discrete, *id_continuous])
    # 1/ detect unnecessary segment
    indexes = [*id_discrete, *id_continuous]
    no_index = list(set(data_.columns).difference(indexes))
    id1, id2 = id_continuous

    disc = tools.compute_discontinuity(data_, id_discrete, id_continuous)
    identical = False * np.ones_like(disc)

    index = data_.index

    data_1 = data_.loc[index[:-1], no_index].fillna(np.nan).values
    data_2 = data_.loc[index[1:], no_index].fillna(np.nan).values

    np_bool: np.array = np.equal(data_1, data_2)

    res = pd.Series(np_bool.sum(axis=1), index=index[:-1])
    res = pd.Series(res == len(no_index)).values

    identical[:-1] = res
    identical[:-1] = identical[:-1] & ~disc[1:]

    n = identical.sum()
    if n == 0:
        return df
    dat = pd.DataFrame(dict(
        identical=identical,
        keep=False * np.ones_like(disc)),
        index=data_.index)

    keep = list(set(np.where(identical)[0]).union(np.where(identical)[0] + 1))
    dat.loc[dat.index[keep], "keep"] = True

    data_merge = data_.sort_values([*id_discrete, *id_continuous])
    data_merge[f"{id1}_new"] = np.nan
    b = ~ dat["identical"]
    b_disc = [True] + list(~dat["identical"].values[:-1])
    data_merge.loc[b, f"{id2}_new"] = data_merge.loc[b, id2]
    data_merge.loc[b_disc, f"{id1}_new"] = data_merge.loc[b_disc, id1]

    data_merge[f"{id2}_new"] = data_merge[f"{id2}_new"].bfill()
    data_merge[f"{id1}_new"] = data_merge[f"{id1}_new"].ffill()

    data_merge = data_merge.drop(list(id_continuous), axis=1)
    data_merge = data_merge.rename({f"{id1}_new": id1, f"{id2}_new": id2},
                                   axis=1)
    return data_merge[df.columns].drop_duplicates().astype(dtypes)


def __merge_index(data_left, data_right,
                  id_discrete,
                  id_continuous,
                  names=("left", "right")):
    id_ = [*id_discrete, *id_continuous]
    id_c = id_continuous
    cr = is_event(data_right, id_continuous=id_continuous)
    cl = is_event(data_left, id_continuous=id_continuous)
    if cr and cl:
        raise AssertionError(
            "[merge] This functionality is not yet implemented")
    elif cl:
        return __merge_index(data_right, data_left, id_discrete=id_discrete,
                             id_continuous=id_c, names=names)
    elif cr:
        data_left = data_left.loc[:, id_].dropna()
        data_left.loc[:, id_c] = data_left.loc[:, id_c].astype(int)

        data_right = data_right.loc[:, [*id_discrete, "pk"]]
        df_merge = __merge_event(data_left, data_right,
                                 id_discrete=id_discrete,
                                 id_continuous=id_c, names=names)
    else:
        data_left = data_left.loc[:, id_].dropna()
        data_right = data_right.loc[:, id_].dropna()
        df_merge = __merge(data_left, data_right,
                           id_discrete=id_discrete, id_continuous=id_c)
    return df_merge


def __merge_event(
        data_left, data_right,
        id_discrete,
        id_continuous,
        names=("left", "right")):
    data_left_ = data_left.iloc[:, :]
    data_right_ = data_right.iloc[:, :]
    data_left_ = _increasing_continuous_index(data_left_, id_continuous)

    index_right = names[1] + "_idx"
    index_left = names[0] + "_idx"
    data_left_.index.name = index_left
    data_right_.index.name = index_right
    data_left_ = data_left_.reset_index()
    data_right_ = data_right_.reset_index()

    id_ = [index_left, *id_discrete, *id_continuous]
    data_left_[index_right] = np.nan
    all_columns = list(set(data_left_.columns).union(data_right_.columns))
    df_merge = data_left_.reindex(columns=all_columns)
    df_merge["__t"] = df_merge[id_continuous[0]]
    data_right_ = data_right_.reindex(columns=all_columns)
    df_merge = pd.concat((df_merge, data_right_), axis=0).sort_values(
        [*id_discrete, "__t"])
    df_merge[id_] = df_merge[id_].fillna(method='pad')
    df_merge = df_merge[~df_merge[index_right].isna()]
    df_merge = df_merge.dropna()
    df_merge["__t"] = df_merge["__t"].fillna(-1).astype(int)
    del data_right_
    del data_left_
    return df_merge


def __merge(df_left: pd.DataFrame, df_right: pd.DataFrame,
            id_discrete: iter,
            id_continuous,
            names=("left", "right")):
    index = [*id_discrete, *id_continuous]

    df_id1, df_id2, index_left, index_right = __refactor_data(
        df_left,
        df_right, id_continuous, id_discrete,
        names=names)
    df_id1_stretched = __fill_stretch(
        df_id1, id_discrete=id_discrete,
        id_continuous=id_continuous)
    df_id2_stretched = __fill_stretch(
        df_id2, id_discrete=id_discrete,
        id_continuous=id_continuous)

    df_id1_stretched.loc[df_id1_stretched[index_left].isna(), index_left] = -1
    df_id2_stretched.loc[df_id2_stretched[index_right].isna(), index_right] = -1

    df = pd.concat((df_id1_stretched, df_id2_stretched), sort=False)
    df = df.sort_values(by=index)

    id1, id2 = id_continuous
    df_merge = __table_jumps(df, *id_continuous, id_discrete)

    df_merge = df_merge.dropna()

    df_merge = pd.merge(
        df_merge,
        df_id1_stretched[[index_left, id1, *id_discrete]],
        on=[id1, *id_discrete], how="left")

    df_merge = pd.merge(
        df_merge,
        df_id2_stretched[[index_right, id1, *id_discrete]],
        on=[id1, *id_discrete], how="left")

    df_end1 = df_id1_stretched[[index_left, id2, *id_discrete]].rename(
        {index_left: index_left + "_end"}, axis=1)
    df_end2 = df_id2_stretched[[index_right, id2, *id_discrete]].rename(
        {index_right: index_right + "_end"}, axis=1)

    df_merge = pd.merge(
        df_merge,
        df_end1,
        left_on=[id1, *id_discrete],
        right_on=[id2, *id_discrete], how="left", suffixes=("", "_1"))
    df_merge = pd.merge(
        df_merge,
        df_end2,
        left_on=[id1, *id_discrete],
        right_on=[id2, *id_discrete], how="left", suffixes=("", "_2"))
    df_merge = df_merge.drop([f"{id2}_1", f"{id2}_2"], axis=1)

    # Tackle the problem of ending pad when there is discontinuity
    idx1 = (df_merge[index_left + "_end"].infer_objects(copy=False).fillna(-1) >= 0).values
    is_na_condition = df_merge.loc[:, index_left].isna()
    df_merge.loc[idx1 & is_na_condition, index_left] = -1

    idx2 = (df_merge[index_right + "_end"].infer_objects(copy=False).fillna(-1) >= 0).values
    is_na_condition_2 = df_merge.loc[:, index_right].isna()
    df_merge.loc[idx2 & is_na_condition_2, index_right] = -1
    df_merge = df_merge.drop([index_right + "_end", index_left + "_end"],
                             axis=1)

    discontinuity = tools.compute_discontinuity(df_merge, id_discrete, id_continuous)
    df_merge.loc[discontinuity & df_merge[
        index_left].isna(), index_left] = -1
    df_merge.loc[discontinuity & df_merge[
        index_right].isna(), index_right] = -1

    df_merge = df_merge.infer_objects(copy=False).ffill().drop("___t", axis=1)

    df_merge[[index_right, index_left, id1, id2]] = df_merge[
        [index_right, index_left, id1, id2]].astype(float).fillna(
        -1).astype(int)

    df_merge = df_merge.loc[
        ~(df_merge[index_left] + df_merge[index_right] == -2)]
    return df_merge


def is_event(data, id_continuous: iter):
    id_continuous = list(id_continuous)
    if id_continuous[0] in data.columns and id_continuous[1] in data.columns:
        return False
    return True


def __fill_stretch(df: pd.DataFrame, id_discrete: iter, id_continuous: iter):
    data1 = df.__deepcopy__()
    data1["added"] = False
    col_save = np.array(data1.columns)
    index = [*id_discrete, *id_continuous]
    data1["discontinuity"] = tools.compute_discontinuity(data1, id_discrete,
                                                   id_continuous)
    ix__ = np.where(data1["discontinuity"].values)[0]
    if data1["discontinuity"].sum() == 0:
        return data1.loc[:, col_save]
    else:
        id1 = id_continuous[0]
        id2 = id_continuous[1]
        data_add = pd.DataFrame(columns=data1.columns, index=range(len(ix__)))
        data_add[index] = data1.iloc[ix__][index].values
        data_add[id1] = data1.iloc[ix__ - 1].loc[:, id2].values
        data_add[id2] = data1.iloc[ix__].loc[:, id1].values
        data_add["added"] = True
        data1 = pd.concat((data1, data_add.dropna(axis=1, how='all')), axis=0)
        data1 = data1[data1[id1] < data1[id2]]
    return data1.loc[:, col_save]


def __fix_discrete_index(
        data_left: pd.DataFrame,
        data_right: pd.DataFrame,
        id_discrete_left: iter,
        id_discrete_right: iter):
    if len(id_discrete_left) < len(id_discrete_right):
        data_right, data_left = __fix_discrete_index(
            data_right, data_left,
            id_discrete_right, id_discrete_left, )
        return data_left, data_right

    df_id_left = data_left.loc[:, id_discrete_left].drop_duplicates()
    df_id_right = data_right.loc[:, id_discrete_right].drop_duplicates()

    id_inter = [id_ for id_ in id_discrete_right if id_ in id_discrete_left]
    id_inter = list(id_inter)
    df_id_right = pd.merge(df_id_left, df_id_right, on=id_inter)
    data_right = pd.merge(df_id_right, data_right, on=id_discrete_right, how="left")
    return data_left, data_right


def suppress_duplicates(df, id_discrete, continuous_index):
    df = df.sort_values([*id_discrete, *continuous_index])
    df_duplicated = df.drop([*id_discrete, *continuous_index], axis=1)
    mat_duplicated = pd.DataFrame(
        df_duplicated.iloc[1:].values == df_duplicated.iloc[
                                         :-1].values)
    id1 = continuous_index[0]
    id2 = continuous_index[1]
    index = mat_duplicated.sum(axis=1) == df_duplicated.shape[1]
    index = np.where(index)[0]
    df1 = df.iloc[index]
    df2 = df.iloc[index + 1]
    idx_replace = df1[id2].values == df2[id1].values
    idx_to_agg = index[idx_replace]
    i_loc = df1.columns.get_loc(id2)
    df.iloc[idx_to_agg, i_loc] = df.iloc[idx_to_agg + 1, i_loc].values
    df = df.drop(df.index[idx_to_agg + 1])
    return df


def _increasing_continuous_index(df: pd.DataFrame, id_continuous):
    id1 = id_continuous[0]
    id2 = id_continuous[1]
    df[f"{id1}_new"] = df.loc[:, [id1, id2]].min(axis=1)
    df[f"{id2}_new"] = df.loc[:, [id1, id2]].max(axis=1)

    df = df.drop([id1, id2], axis=1)
    df = df.rename({f"{id1}_new": id1, f"{id2}_new": id2}, axis=1)
    return df


def __refactor_data(data_left, data_right, id_continuous, id_discrete,
                    names=("left", "right")):
    index = [*id_discrete, *id_continuous]
    data_left = _increasing_continuous_index(data_left, id_continuous)
    data_right = _increasing_continuous_index(data_right, id_continuous)
    index_right = names[1] + "_idx"
    index_left = names[0] + "_idx"
    df_id1 = data_left[index].drop_duplicates()
    df_id2 = data_right[index].drop_duplicates()
    df_id1.index.name = index_left
    df_id2.index.name = index_right
    df_id1 = df_id1.reset_index()
    df_id2 = df_id2.reset_index()

    df_id1[index_right] = np.nan
    df_id2[index_left] = np.nan

    df_id1 = df_id1.sort_values(by=index)
    df_id2 = df_id2.sort_values(by=index)
    return df_id1, df_id2, index_left, index_right


def __check_args_merge(data_left, data_right,
                       id_continuous,
                       id_discrete,
                       how):
    for c in [*id_continuous, *id_discrete]:
        if not (c in data_left.columns or c in data_right.columns):
            raise ValueError(f"{c} is not in columns")
    if not len(id_continuous) == 2:
        raise ValueError("Only two continuous index is possible")
    if how not in ["left", "right", "inner", "outer"]:
        raise ValueError('How must be in "left", "right", "inner", "outer"')


def __table_jumps(data, id1, id2, id_discrete):
    df_unique_start = data[[id1, *id_discrete]].rename(
        {id1: "___t"}, axis=1).drop_duplicates()

    df_unique_end = data[[id2, *id_discrete]].rename(
        {id2: "___t"}, axis=1).drop_duplicates()

    ret = pd.concat((df_unique_end, df_unique_start),
                    sort=False).sort_values(
        by=[*id_discrete, "___t"]).drop_duplicates()
    ret.index = range(ret.shape[0])
    ret[id1] = -1
    ret.iloc[:-1, -1] = ret["___t"].iloc[:-1].values
    ret[id2] = -1
    ret.iloc[:-1, -1] = ret["___t"].iloc[1:].values
    return ret





# Copyright 2023 Eurobios
# Licensed under the CeCILL License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     https://cecill.info/
import numpy as np
import pandas as pd

from crep import merge, aggregate_constant
from crep import tools
from crep.base import __fill_stretch, __merge


def test_merge_basic(get_examples):
    dfl, dfr = get_examples
    ret = merge(dfl, dfr,
                id_continuous=["t1", "t2"],
                id_discrete=["id"],
                how="outer")
    ret_l = merge(dfl, dfr,
                  id_continuous=["t1", "t2"],
                  id_discrete=["id"],
                  how="left")
    ret_i = merge(dfl, dfr,
                  id_continuous=["t1", "t2"],
                  id_discrete=["id"],
                  how="inner")
    ret_r = merge(dfl, dfr,
                  id_continuous=["t1", "t2"],
                  id_discrete=["id"],
                  how="right", verbose=True)
    ret_th = pd.DataFrame(
        dict(id=[1, 1, 1, 1, 2, 2, 2],
             t1=[0, 5, 10, 80, 0, 100, 120],
             t2=[5, 10, 80, 100, 90, 110, 130],
             data1=[0.2, 0.2, 0.2, 0.2, 0.1, 0.3, 0.2],
             data2=[np.nan, 0.2, 0.2, np.nan, 0.1, 0.3, 0.2],
             ))
    ret_th = ret_th.astype(ret.dtypes)
    ret_i_th = ret_th.dropna()
    ret_i_th.index = range(ret_i_th.__len__())
    assert ret.equals(ret_th)
    assert ret_l.equals(ret_th)
    assert ret_i.equals(ret_i_th)
    assert ret_r.equals(ret_i_th)


def test_fill_stretch(get_examples):
    dfl, _ = get_examples
    ret = __fill_stretch(dfl.__deepcopy__(),
                         id_continuous=["t1", "t2"],
                         id_discrete=["id"],
                         )
    assert ret["added"].sum() == 2


def test__merge(get_advanced_examples):
    df_left, df_right = get_advanced_examples

    df_merge = __merge(df_left, df_right,
                       id_discrete=["id"],
                       id_continuous=["t1", "t2"])


def test_admissible_data(get_advanced_examples):
    df_left, df_right = get_advanced_examples


def test_check_args(get_examples):
    pass


def test_aggregate_constant(get_examples):
    df1, _ = get_examples
    df1.index = np.random.uniform(size=len(df1))
    ret = aggregate_constant(df1, id_continuous=["t1", "t2"],
                             id_discrete=["id"])
    df_left = pd.DataFrame(
        dict(id=[1, 2, 2, 2],
             t1=[0, 0, 100, 120],
             t2=[100, 90, 110, 130],
             data1=[0.2, 0.1, 0.3, 0.2])
    )
    df_left.index = ret.index
    assert len(ret) < len(df1)
    assert df_left.equals(ret)


def test_merge_duplicates(get_examples):
    dfl, dfr = get_examples
    ret_r = merge(dfl, dfr,
                  id_continuous=["t1", "t2"],
                  id_discrete=["id"],
                  how="right", verbose=True, remove_duplicates=True)
    assert all(ret_r.isna().sum() == 0)


def test_aggregate_constant_no_aggregation(get_examples):
    _, dfr = get_examples
    ret = aggregate_constant(dfr, id_continuous=["t1", "t2"],
                             id_discrete=["id"])

    ret2 = aggregate_constant(ret, id_continuous=["t1", "t2"],
                              id_discrete=["id"])
    assert ret.equals(ret2)


def test_merge_symetry(get_examples):
    dfl, dfr = get_examples
    ret1 = merge(dfl, dfr,
                 id_continuous=["t1", "t2"],
                 id_discrete=["id"],
                 how="outer")
    ret2 = merge(dfr, dfl,
                 id_continuous=["t1", "t2"],
                 id_discrete=["id"],
                 how="outer")

    assert ret1.equals(ret2[ret1.columns])


def test_merge_discrete_id(get_advanced_examples):
    data_left, data_right = get_advanced_examples

    ret1 = merge(data_left, data_right,
                 id_continuous=["t1", "t2"],
                 id_discrete=["id", "id2"],
                 how="outer")
    ret2 = merge(data_left, data_right,
                 id_continuous=["t1", "t2"],
                 id_discrete=["id"],
                 how="outer")
    assert len(ret1) > len(ret2)


def test_build_admissible_dataset(get_advanced_examples):
    df_in = get_advanced_examples[1]
    data = {
        'id': [1, 2, 2, 3, 2, 2, 2, 2, 3, 3, 3, 3],
        't1': [5, 100, 120, 30, 0, 10, 10, 80, 10, 20, 20, 25],
        't2': [10, 110, 130, 35, 10, 80, 80, 90, 20, 25, 25, 30],
        'data2': [0.2, 0.3, 0.2, 0.1, 0.1, 0.2, 0.1, 0.1, 0.3, 0.3, 0.2, 0.3]
    }

    # Creating the DataFrame
    df = pd.DataFrame(data)

    df_out = tools.build_admissible_data(df_in, id_continuous=["t1", "t2"],
                                         id_discrete=["id"])
    df_out = df_out.sort_values(by=["id", "t1", "t2"])
    df = df.sort_values(by=["id", "t1", "t2"])

    assert all(df["data2"].values == df_out["data2"].values)




# Copyright 2023 Eurobios
# Licensed under the CeCILL License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     https://cecill.info/

import pandas as pd
import numpy as np

from crep.tools import get_overlapping, admissible_dataframe, sample_non_admissible_data, create_zones, \
    build_admissible_data

id_discrete, id_continuous = ["id", "id2"], ["t1", "t2"]
data = {
    'id': [1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, np.nan],
    'id2': [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
    't1': [932, 996, 2395, 2395, 3033, 3628, 4126, 4140, 4154, 63, 63, 271, 271, np.nan],
    't2': [2395, 2324, 3033, 3628, 3035, 4140, 4140, 5508, 5354, 199, 199, 351, 357, np.nan],
    'LONGUEUR': [1463, 1328, 638, 1233, 2, 512, 14, 1368, 1200, 136, 208, 935, 1086, np.nan],
    '__zone__': [0, 0, 1, 1, 1, 2, 2, 2, 2, 3, 3, 4, 4, 5]
}
data = pd.DataFrame(data).sort_values(by=[*id_discrete, *id_continuous])


def test_no_overlapping(get_examples):
    assert sum(get_overlapping(
        get_examples[0], ["id"], ["t1", "t2"]
    )) == 0
    assert admissible_dataframe(get_examples[0], ["id"], id_continuous)


def test_overlapping(get_examples):
    df = get_examples[0]
    df.loc[1, "t1"] = 5
    ret = get_overlapping(df, ["id"], id_continuous)
    assert sum(ret) == 2


def test_sample_overlapping(get_examples):
    df = get_examples[0]
    df.loc[1, "t1"] = 5
    ret = sample_non_admissible_data(df, ["id"], id_continuous)
    assert ret.equals(df.loc[[0, 1]])


def test_create_zones():
    df = data.sort_values(by=[*id_discrete, *id_continuous])
    df_out = create_zones(df.drop(columns="__zone__"), id_discrete=id_discrete, id_continuous=id_continuous)
    df_out = df_out.sort_values(by=[*id_discrete, *id_continuous])
    assert all(df_out["__zone__"].values == df["__zone__"].values)


def test_build_admissible_data_frame():
    ret = build_admissible_data(df=data.drop(columns="__zone__"), id_discrete=id_discrete, id_continuous=id_continuous)
    ret = ret.drop_duplicates(subset=[*id_discrete, *id_continuous])

    assert admissible_dataframe(ret, id_discrete=id_discrete, id_continuous=id_continuous)


def test_fix_point_non_admissible(get_examples):

    df_non_admissible = sample_non_admissible_data(data, id_discrete, id_continuous)
    df_non_admissible_2 = sample_non_admissible_data(df_non_admissible, id_discrete, id_continuous)

    assert df_non_admissible.equals(df_non_admissible_2)

    df_non_admissible = sample_non_admissible_data(get_examples[0], ['id'], id_continuous)
    df_non_admissible_2 = sample_non_admissible_data(df_non_admissible, ['id'], id_continuous)

    assert df_non_admissible.equals(df_non_admissible_2)

    df_non_admissible = sample_non_admissible_data(get_examples[1], ['id'], id_continuous)
    df_non_admissible_2 = sample_non_admissible_data(df_non_admissible, ['id'], id_continuous)

    assert df_non_admissible.equals(df_non_admissible_2)
